import numpy as np
import pandas as pd

from epidemic import *

# Load our training data
df_infectados = pd.read_csv('data/chile_minsal_infectados.csv', sep=';', index_col=0)
df_infectados = df_infectados.transpose()
df_infectados.index = pd.to_datetime(df_infectados.index, format='%d-%b') + pd.offsets.DateOffset(years=120)

df_fallecidos = pd.read_csv('data/chile_minsal_fallecidos.csv', sep=';', index_col=0)
df_fallecidos = df_fallecidos.transpose()
df_fallecidos.index = pd.to_datetime(df_fallecidos.index, format='%d-%b') + pd.offsets.DateOffset(years=120)

data = pd.DataFrame({
    'I_cases': df_infectados['Región Metropolitana'],
    'D': df_fallecidos['Región Metropolitana'],
})

# Initial time and state
t0 = np.datetime64('2020-04-04')

I0 = data['I_cases'].loc[t0]
D0 = data['D'].loc[t0]

# Define our training parameters: initial value, bounds, and mapping function to model parameters
x = [
    0.74,   # E0
    10.3,   # Im0
    0.38,   # CE
    0.75,   # CIm
    0.165,  # βI
    0.2,    # γE
    0.1,    # γIm
    0.1,    # γI
    0.1667, # γH
    0.1,    # γHc
]

x_bounds = [
    (0,20),        # E0
    (0,20),        # Im0
    (0.0,0.4),     # CE
    (0.0,0.9),     # CIm
    (0.0,0.75),    # βI
    (0.17,0.25),   # γE
    (0.07,0.14),   # γIm
    (0.07,0.14),   # γI
    (0.1,0.5),     # γH
    (0.0625,0.14), # γHc
]

def x_params(E0, Im0, CE, CIm, βI, γE, γIm, γI, γH, γHc):
    y0 = {
        'S': 1e7,
        'E': E0 * I0,
        'Im': Im0 * I0,
        'I': I0,
        'H': 0,
        'Hc': 0,
        'R': 0,
        'D': D0,
    }
    
    λ1 = np.datetime64('2020-04-15')
    κ1 = 0.05
    α2 = 0.75
    α = lambda t: 1.0 if t < λ1 else α2 + (1.0-α2)*np.exp(-κ1*(t-λ1)/np.timedelta64(1,'D'))
    return y0, lambda t: {
        'βE': α(t)*CE*βI,
        'βIm': α(t)*CIm*βI,
        'βI': α(t)*βI,
        'βH': 0.0,
        'βHc': 0.0,
        'γE': γE,
        'γIm': γIm,
        'γI': γI,
        'γH': γH,
        'γHc': γHc,
        'μb': 3.57e-5,
        'μd': 1.57e-5,
        'φEI': 0.5,
        'φIR': 0.6,
        'φHR': 0.6,
        'φD': 0.2,
    }

# Create model till time 'tmax'
model = ModelReport2()
epidemic = Epidemic(model, t0, tmax=np.datetime64('2020-06-01'), data=data)

# Optimize using dual_annealing and L-BFGS-B subsequently
options = {
    'annealing': {
        'seed': 1234567,
        'fast': True,
    },
    'L-BFGS-B': {
        'disp': True,
    },
}

for method in ['annealing', 'L-BFGS-B']:
    opt = {}
    if method in options:
        opt = options[method]
    x = epidemic.optimize(x, x_bounds, x_params, method=method, **opt)

epidemic.plot(cols=['I_cases', 'I'])
epidemic.plot(cols=['D'])



import inspect
import theano
import numpy as np

from .epidemia import Model

if not 'profile' in dir():
    def profile(f):
        return f

class SIR(Model):
    def __init__(self):
        self.cols = ['S', 'I', 'R',
                     'I_cases']

        self.names = {
            'I_cases': r'$I_{\mathrm{cases}}$',
        }
        self.styles = {
            'S': '-.', 'I': '-.', 'R': '-.',
        }
        self.colors = {
            'S': 'tab:gray', 'I': 'tab:blue', 'R': 'tab:green',
            'I_cases': 'tab:blue',
        }

        Model.__init__(self)

    def first(self, y0):
        if 'I_cases' not in y0:
            y0['I_cases'] = y0['I']
        return [y0[col] for col in self.cols]
    
    @profile
    def step(self, t, y, params):
        p = params(t)
        S, I, R = y[0], y[1], y[2]
        β, γ, μb, μd = p['β'], p['γ'], p['μb'], p['μd']

        N = S + I + R

        return [
            # S, I, R
            μb*N - β*I*S/N - μd*S,
            β*I*S/N - γ*I,
            γ*I,

            # I_cases
            β*I*S/N,
        ]

    def N(self, t, Y=None, params=None):
        if Y is None:
            Y = self.Y
        y = Y[t]
        return y['S'] + y['I'] + y['R']
    
    def R(self, t, Y=None, params=None):
        if params is None:
            params = self.params
        p = params(t)
        return p['β'] / p['γ']
    
    def R_effective(self, t, Y=None, params=None):
        if Y is None:
            Y = self.Y
        y = Y[t]
        return self.R(t, Y, params) * y['S'] / self.N(t, Y, params)

class ModelReport2(Model):
    def __init__(self):
        self.cols = ['S', 'E', 'Im', 'I', 'H', 'Hc', 'R', 'D',
                     'Im_cases', 'I_cases', 'H_cases', 'Hc_cases', 'R_cases', 'H_total']

        self.names = {
            'Im_cases': r'$I^m_{\mathrm{cases}}$', 'I_cases': r'$I_{\mathrm{cases}}$', 'H_cases': r'$H_{\mathrm{cases}}$',
            'Hc_cases': r'$H^c_{\mathrm{cases}}$', 'R_cases': r'$R_{\mathrm{cases}}$', 'H_total': r'$H + H^c$',
            'R_effective': '$R_{eff}$', 'Fatality': r'$\mathrm{Fatality}$',
        }
        self.styles = {
            'S': '-.', 'E': '-.', 'Im': '-.', 'I': '-.',
            'H': '-.', 'Hc': '-.', 'R': '-.', 'D': '-.',
            'H_total': '-.',
        }
        self.colors = {
            'S': 'tab:gray', 'E': 'tab:cyan', 'Im': 'tab:purple', 'I': 'tab:blue',
            'H': 'tab:olive', 'Hc': 'tab:orange', 'R': 'tab:green', 'D': 'tab:red',
            'Im_cases': 'tab:purple', 'I_cases': 'tab:blue', 'H_cases': 'tab:olive',
            'Hc_cases': 'tab:orange', 'R_cases': 'tab:green', 'H_total': 'tab:brown',
        }

        Model.__init__(self)

    def first(self, y0):
        if 'Im_cases' not in y0:
            y0['Im_cases'] = y0['Im']
        if 'I_cases' not in y0:
            y0['I_cases'] = y0['I']
        if 'H_cases' not in y0:
            y0['H_cases'] = y0['H']
        if 'Hc_cases' not in y0:
            y0['Hc_cases'] = y0['Hc']
        if 'R_cases' not in y0:
            y0['R_cases'] = y0['R']
        if 'H_total' not in y0:
            y0['H_total'] = y0['H'] + y0['Hc']
        return [y0[col] for col in self.cols]
    
    @profile
    def step(self, t, y, params):
        p = params(t)
        S, E, Im, I, H, Hc, R = y[0], y[1], y[2], y[3], y[4], y[5], y[6]
        βE, βIm, βI, βH, βHc = p['βE'], p['βIm'], p['βI'], p['βH'], p['βHc']
        γE, γIm, γI, γH, γHc = p['γE'], p['γIm'], p['γI'], p['γH'], p['γHc']
        μb, μd = p['μb'], p['μd']
        φEI, φIR, φHR, φD = p['φEI'], p['φIR'], p['φHR'], p['φD']

        N = S + E + Im + I + H + Hc + R
        ROC = (1.0/N) * (βE*E + βIm*Im + βI*I + βH*H + βHc*Hc)

        return [
            # S, E, Im, I, H, Hc, R, D
            μb*N - S*ROC - μd*S,
            S*ROC - (γE+μd)*E,
            (1.0-φEI)*γE*E - (γIm+μd)*Im,
            φEI*γE*E - (γI+μd)*I,
            (1.0-φIR)*γI*I + (1-φD)*γHc*Hc - (γH+μd)*H,
            (1.0-φHR)*γH*H - (γHc+μd)*Hc,
            γIm*Im + φIR*γI*I + φHR*γH*H - μd*R,
            φD*γHc*Hc,
            
            # Im_cases, I_cases, H_cases, Hc_cases, R_cases
            (1.0-φEI)*γE*E,
            φEI*γE*E,
            (1.0-φIR)*γI*I,
            (1.0-φHR)*γH*H,
            φIR*γI*I + φHR*γH*H,

            # H_total
            (1.0-φIR)*γI*I + (1-φD)*γHc*Hc - (γH+μd)*H + (1.0-φHR)*γH*H - (γHc+μd)*Hc,
        ]

    def N(self, t, Y=None, params=None):
        if Y is None:
            Y = self.Y
        y = Y[t]
        return y['S'] + y['E'] + y['Im'] + y['I'] + y['H'] + y['Hc'] + y['R']
    
    def Fatality(self, t, Y=None, params=None):
        if Y is None:
            Y = self.Y
        y = Y[t]
        if y['I_cases'] == 0:
            return 0
        return y['D'] / y['I_cases']
    
    def R(self, t, Y=None, params=None):
        if params is None:
            params = self.params
        p = params(t)

        # see O. Diekmann, J.A.P. Heesterbeek, and M.G. Roberts, "The construction of next-generation matrices for compartmental epidemic models", J R Soc Interface, 2010
        # T are transmissions
        # Σ are transitions
        Σ = np.matrix([
            [-p['γE']-p['μd'], 0, 0, 0, 0],
            [(1.0-p['φEI'])*p['γE'], -p['γIm']-p['μd'], 0, 0, 0],
            [p['φEI']*p['γE'], 0, -p['γI']-p['μd'], 0, 0],
            [0, 0, (1.0-p['φIR'])*p['γI'], -p['γH']-p['μd'], (1.0-p['φD'])*p['γHc']],
            [0, 0, 0, (1.0-p['φHR'])*p['γH'], -p['γHc']-p['μd']],
        ])
        T = np.matrix([
            [p['βE'], p['βIm'], p['βI'], p['βH'], p['βHc']],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
        ])
        
        # K = -T * Σ^-1  =>  (Σ.T) * (K.T) = -T.T
        K = np.linalg.solve(Σ.T, -T.T).T
        if np.linalg.det(K) != 0.0:
            logging.warn('det(K) = %g != 0' % np.linalg.det(K))
        return K.trace()[0,0]
    
    def R_effective(self, t, Y=None, params=None):
        if Y is None:
            Y = self.Y
        y = Y[t]
        return self.R(t, Y, params) * y['S'] / self.N(t, Y, params)



import pymc3
import theano
import numpy as np
import scipy

class DifferentialEquation(pymc3.ode.DifferentialEquation):
    def __init__(self, func, times, *, n_states, n_theta, t0=0, model, x_params, options={'method': 'RK45'}):
        def func2(y, t, p):
            ps = [p[i] for i in range(n_theta)]
            return func(t, y, x_params(*ps)[1])

        if times[0] == t0:
            times = times[1:]

        pymc3.ode.DifferentialEquation.__init__(self, func2, times, n_states=n_states, n_theta=n_theta, t0=t0)

        self.model = model
        self.x_params = x_params
        self.options = options

    def _system(self, t, Y, p):
        r"""This is the function that will be passed to odeint. Solves both ODE and sensitivities.
        Parameters
        ----------
        t : float
            current time
        Y : array
            augmented state vector (n_states + n_states + n_theta)
        p : array
            parameter vector (y0, theta)
        """
        dydt, ddt_dydp = self._augmented_func(Y[:self.n_states], t, p, Y[self.n_states:])
        derivatives = np.concatenate([dydt, ddt_dydp])
        return derivatives

    def _simulate(self, y0, theta):
        y0, _ = self.x_params(*theta)
        y0 = self.model.first(y0)

        # Initial condition comprised of state initial conditions and raveled sensitivity matrix
        s0 = np.concatenate([y0, self._sens_ic])

        # perform the integration
        res = scipy.integrate.solve_ivp(self._system, (self._augmented_times[0], self._augmented_times[-1]),
                s0, t_eval=self._augmented_times, args=(np.concatenate([y0, theta]),), **self.options)
        sol = res['y'].T.astype(theano.config.floatX)

        # The solution
        y = sol[1:, :self.n_states]

        # The sensitivities, reshaped to be a sequence of matrices
        sens = sol[1:, self.n_states:].reshape(self.n_times, self.n_states, self.n_p)

        return y, sens




df_confirmed = pd.read_csv('data_confirmed.csv')
df_confirmed.drop(['Province/State', 'Lat', 'Long'], axis=1, inplace=True)
df_confirmed = df_confirmed.groupby('Country/Region').sum()
df_confirmed = df_confirmed.transpose()
df_confirmed.index = pd.to_datetime(df_confirmed.index)

df_recovered = pd.read_csv('data_recovered.csv')
df_recovered.drop(['Province/State', 'Lat', 'Long'], axis=1, inplace=True)
df_recovered = df_recovered.groupby('Country/Region').sum()
df_recovered = df_recovered.transpose()
df_recovered.index = pd.to_datetime(df_recovered.index)

df_deaths = pd.read_csv('data_deaths.csv')
df_deaths.drop(['Province/State', 'Lat', 'Long'], axis=1, inplace=True)
df_deaths = df_deaths.groupby('Country/Region').sum()
df_deaths = df_deaths.transpose()
df_deaths.index = pd.to_datetime(df_deaths.index)

I = df_confirmed['China'].copy()
R = df_recovered['China'].copy()
D = df_deaths['China'].copy()

# adjust infection data
idx_adj = pd.date_range(I.index.min(), '2020-02-13', freq='D')
data_adj = [17410*(I[I.index.min():t].sum()/I[idx_adj].sum()) for t in idx_adj]
I_adj = pd.Series(data_adj, index=idx_adj)
I_adj['2020-02-13'] -= 14000
I = I.add(I_adj, fill_value=0)

#jan24 = np.datetime64('2020-01-24')
#feb08 = np.datetime64('2020-02-08')
#θ = lambda t: 0.14 if t < jan24 else (
#              0.65 if t >= feb08 else (
#              0.14 + (0.65-0.14)*((t-jan24)/np.timedelta64(1,'D')/16)))

data = pd.DataFrame({'I_cases': df_confirmed['China'], 'R_cases': df_recovered['China'], 'D_cases': df_deaths['China']})


ivorra21M_x0 = [0.3373, 0.8450, 0.3622, 7.0004, 0.1815, 0.0050, 0.1200, 0.9515, 0.665]
ivorra25F_x0 = [0.2771, 0.9426, 0.6336, 13.9, 0.1070, 0.0458, 0.1272, 0.9515, 0.665]
ivorra8F_x0 = [0.4526, 0.2665, 0.9255, 13.2, 0.1476, 0.0330, 0.1647, 0.9515, 0.665]
ivorra29J_x0 = [0.5000, 0.1543, 0.9988, 13.7, 0.1128, 0.0500, 0.2354, 0.9515, 0.665]

ivorra_x_bounds = [(1e-6,0.5), (0.0,1.0), (0.0,1.0), (7.0,14.0),
                   (0.0,0.2), (0.005,0.05), (0.0,1.0),
                   (0.0,1.0), (0.0,1.0)]

def ivorra_x_params(βI, CE, CIm, δR, δω, ωc, κ1, φIR, φHR):
    # notation differences with Ivorra paper:
    # βIu, γInf, dIu => βIm, γIm, dIm
    # βId, dId => βI, dI
    # βHR, γHR => βH, γH
    # βHD, γHD => βHc, γHc
    # ω top bar => ωnc
    # ω bottom bar => ωc
    # φIR and φHR are not known in Ivorra
    
    λ1 = np.datetime64('2020-02-01')
    m = lambda t: 1.0 if t < λ1 else np.exp(-κ1*(t-λ1)/np.timedelta64(1,'D'))
    
    ωnc = ωc + δω
    ω = lambda t: m(t)*ωnc + (1-m(t))*ωc
    
    θt0 = np.datetime64('2020-01-24')
    θt1 = np.datetime64('2020-02-08')
    θ = lambda t: 0.86 if t < θt0 else (0.35 if t > θt1 else 0.86+(0.35-0.86)*(t-θt0)/(θt1-θt0))
        
    dg = 5.7
    dE = 5.5
    dI = 6.7
    dIm = 14
    dH = (dIm-dI)
    g = lambda t: dg*(1-m(t))
    γE = 1/ dE
    γIm = 1/ dIm
    γI = lambda t: 1 / (dI - g(t))
    γH = lambda t: 1 / (dIm - dI + g(t))
    γHc = 1 / (dH + δR)    
    
    βE = CE*βI
    βIm = CIm*βI
    CH = lambda t: 0.0275*(βI/γI(t) + βE/γE)/((1-0.0275)*βI*((1-ω(t))/γH(t) + ω(t)/γHc))
    βH = lambda t: CH(t)*βI
    βHc = lambda t: CH(t)*βI
            
    return lambda t: {'βE': m(t)*βE, 
                      'βIm': m(t)*βIm,
                      'βI': m(t)*βI,
                      'βH': m(t)*βH(t),
                      'βHc': m(t)*βHc(t),
                      'γE': γE,
                      'γIm': γIm,
                      'γI': γI(t),
                      'γH': γH(t),
                      'γHc': γHc,
                      'μb': 0.0,
                      'μd': 0.0,
                      'φEI': θ(t),
                      'φIR': φIR,
                      'φHR': φHR,
                      'φD': ω(t)}
